﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter1
{
    class Item01
    {
        public class Customer
        {
            private string name;
            public string Name
            {
                get => name;
                set
                {
                    if (string.IsNullOrEmpty(value))
                        throw new ArgumentException(
                            "Name cannot be blank",
                          "Name");
                    name = value;
                }
                // More Elided.
            }
        }

        private class Nested
        {
            internal void Add(String v1, String address, String v2)
            {
                throw new NotImplementedException();
            }
        }
        private class Sample
        {
            public Nested DataBindings => default(Nested);
        }
        private static void Snippet01()
        {
            var textBoxCity = new Sample();
            var address = "txt";
            textBoxCity.DataBindings.Add("Text",
                address, "City");

        }
        public static void Tests()
        {

        }
    }
    namespace Sample2
    {
        public class Customer
        {
            private object syncHandle = new object();

            private string name;
            public string Name
            {
                get
                {
                    lock (syncHandle)
                        return name;
                }
                set
                {
                    if (string.IsNullOrEmpty(value))
                        throw new ArgumentException(
                            "Name cannot be blank",
                          "Name");
                    lock (syncHandle)
                        name = value;
                }
            }
            // More Elided.
        }

    }

    namespace Sample3
    {
        public class Customer
        {
            public virtual string Name
            {
                get;
                set;
            }
        }

    }

    namespace Sample4
    {
        public interface INameValuePair<T>
        {
            string Name
            {
                get;
            }

            T Value
            {
                get;
                set;
            }
        }

        public class Customer
        {
            public virtual string Name
            {
                get;
                protected set;
            }
            // remaining implementation omitted
        }

        class Sample
        {
            private int[] theValues = new int[100];
            public int this[int index]
            {
                get => theValues[index];
                set => theValues[index] = value;
            }

            public static void Usage()
            {
                int i = 5;
                int[] someObject = new int[20];
                // Accessing an indexer:
                int val = someObject[i];
            }
        }
    }

    namespace Sample5
    {
        public class Unused
        {
            private Dictionary<string, Address> addressValues;
            public Address this[string name]
            {
                get => addressValues[name];
                set => addressValues[name] = value;
            }
        }

        public class Address
        {
            public int this[int x, int y]
                => ComputeValue(x, y);

            public int this[int x, string name]
                => ComputeValue(x, name);

            private Int32 ComputeValue(Int32 x, String name)
            {
                throw new NotImplementedException();
            }
            private Int32 ComputeValue(Int32 x, Int32 y)
            {
                throw new NotImplementedException();
            }

        }
        // using public data members, bad practice:
        public class Customer
        {
            public string Name;

            // remaining implementation omitted
        }

        public class Sample
        {
            private void Method()
            {
                var customerOne = new Customer();
                string name = customerOne.Name;
                customerOne.Name = "This Company, Inc.";

            }
        }

    }
}
