﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter1
{
    class Item02
    {
        private class Snippet01
        {
            public string Name
            {
                get;
                set;
            }
        }
        private class Snippet02
        {
            public string Name
            {
                get;
                protected set;
            }
            // Or
            public string Name2
            {
                get;
                internal set;
            }
            // Or
            public string Name3
            {
                get;
                protected internal set;
            }
            // Or
            public string Name4
            {
                get;
                private set;
            }
            public string Name5
            {
                get;
            }
        }

        public class Sample3
        {
            public class BaseType
            {
                public virtual string Name
                {
                    get;
                    protected set;
                }
            }

            public class DerivedType : BaseType
            {
                public override string Name
                {
                    get => base.Name;
                    protected set
                    {
                        if (!string.IsNullOrEmpty(value))
                            base.Name = value;
                    }
                }
            }

        }

        class Sample4
        {

            // original version
            public class Person
            {
                public string FirstName
                {
                    get;
                    set;
                }
                public string LastName
                {
                    get;
                    set;
                }
                public override string ToString() =>
                    $"{FirstName} {LastName}";
            }
        }

        class Sample5
        {
            // Later updated for validation
            public class Person
            {
                public Person(string firstName, string lastName)
                {
                    // leverage validation in property setters:
                    this.FirstName = firstName;
                    this.LastName = lastName;
                }
                private string firstName;
                public string FirstName
                {
                    get => firstName;
                    set
                    {
                        if (string.IsNullOrEmpty(value))
                            throw new ArgumentException(
                                "First name cannot be null or empty");
                        firstName = value;
                    }
                }

                private string lastName;
                public string LastName
                {
                    get => lastName;
                    private set
                    {
                        if (string.IsNullOrEmpty(value))
                            throw new ArgumentException(
                                "Last name cannot be null or empty");
                        lastName = value;
                    }
                }
                public override string ToString() =>
                    $"{FirstName} {LastName}";
            }
        }

        public static void Tests()
        {

        }
    }
}
