﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter2
{
    public static class Extensions
    {
        public static void ForAll<T>(
            this IEnumerable<T> sequence,
            Action<T> action)
        {
            foreach (T item in sequence)
                action(item);
        }
    }

    class Item14
    {
        public static void PrintCollection<T>(
            IEnumerable<T> collection)
        {
            foreach (T o in collection)
                Console.WriteLine($"Collection contains {o}");
        }

        public static void PrintCollection(
            System.Collections.IEnumerable collection)
        {
            foreach (object o in collection)
                Console.WriteLine($"Collection contains {o}");
        }

        public static void PrintCollection(
            WeatherDataStream collection)
        {
            foreach (object o in collection)
                Console.WriteLine($"Collection contains {o}");
        }

        public static void Tests()
        {
            var foo = new List<int>();
            // usage
            foo.ForAll((n) => Console.WriteLine(n.ToString()));

            var warmDays = from item in
                               new WeatherDataStream("Ann Arbor")
                           where item.Temperature > 80
                           select item;

            var warmDays2 = new WeatherDataStream("Ann Arbor").
                Where(item => item.Temperature > 80);

            // Don't write this, for explanatory purposes
            var warmDays3 = Enumerable.Select(
                                Enumerable.Where(
                                new WeatherDataStream("Ann Arbor"),
                                item => item.Temperature > 80),
                                item => item);


        }
    }

    public enum Direction
    {
        North,
        NorthEast,
        East,
        SouthEast,
        South,
        SouthWest,
        West,
        NorthWest
    }

    public class WeatherData
    {
        public WeatherData(double temp, int speed, Direction direction)
        {
            Temperature = temp;
            WindSpeed = speed;
            WindDirection = direction;
        }
        public double Temperature { get;  }
        public int WindSpeed { get;  }
        public Direction WindDirection { get;  }
        public override string ToString() =>
            $"Temperature = {Temperature}, Wind is {WindSpeed} mph from the {WindDirection}";
    }

    public class WeatherDataStream : IEnumerable<WeatherData>
    {
        private Random generator = new Random();

        public WeatherDataStream(string location)
        {
            // elided
        }

        private IEnumerator<WeatherData> getElements()
        {
            // Real implementation would read from
            // a weather station.
            for (int i = 0; i < 100; i++)
                yield return new WeatherData(
                    temp: generator.NextDouble() * 90,
                    speed: generator.Next(70),
                    direction: (Direction)generator.Next(7)
                );
        }

        public IEnumerator<WeatherData> GetEnumerator() =>
            getElements();

        System.Collections.IEnumerator
            System.Collections.IEnumerable.GetEnumerator() =>
            getElements();

        public List<WeatherData> DataSequence => sequence; 
        private List<WeatherData> sequence = new List<WeatherData>();

    }

    namespace LaterExample
    {
        public class Employee
        {
            public string FirstName { get; set; }
            public string LastName { get; set; }

            public string Name => $"{LastName}, {FirstName}";

            // other details elided.
        }

        public class Customer
        {
            public string Name => customerName;

            // other details elided
            private string customerName;
        }

        public class Vendor
        {
            public string Name => vendorName;

            // other details elided
            private string vendorName;
        }
    }
    namespace Sample3
    {
        public class PhoneNumber { }
        public class Address { }

        public interface IContactInfo
        {
            string Name { get; }
            PhoneNumber PrimaryContact { get; }
            PhoneNumber Fax { get; }
            Address PrimaryAddress { get; }
        }

        public class Employee : IContactInfo
        {
            // implementation elided.
            public PhoneNumber Fax
            {
                get
                {
                    throw new NotImplementedException();
                }
            }

            public string Name
            {
                get
                {
                    throw new NotImplementedException();
                }
            }

            public Address PrimaryAddress
            {
                get
                {
                    throw new NotImplementedException();
                }
            }

            public PhoneNumber PrimaryContact
            {
                get
                {
                    throw new NotImplementedException();
                }
            }
        }

        class Usage
        {
            public void PrintMailingLabel(IContactInfo ic)
            {
                // implementation deleted.
            }

            public struct URLInfo : IComparable<URLInfo>, IComparable
            {
                private Uri URL;
                private string description;

                // Compare the string representation of 
                // the URL:
                public int CompareTo(URLInfo other) => 
                    URL.ToString().CompareTo(other.URL.ToString());

                int IComparable.CompareTo(object obj) =>
                    (obj is URLInfo other) ? 
                        CompareTo(other) :
                        throw new ArgumentException(
                          message: "Compared object is not URLInfo",
                          paramName: nameof(obj));
            }

        }
    }
}
