﻿using System;
using static System.Console;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter2
{
    interface IMessage
    {
        void Message();
    }

    public class MyClass : IMessage
    {
        public void Message() =>
            WriteLine(nameof(MyClass));
    }

    public class MyDerivedClass : MyClass
    {
        public new void Message() =>
            WriteLine(nameof(MyDerivedClass));
    }


    public class Item15
    {
        public static void Tests()
        {
            MyDerivedClass d = new MyDerivedClass();
            d.Message(); // prints "MyDerivedClass".
            IMessage m = d as IMessage;
            m.Message(); // prints "MyClass"
        }
        
        public static void Test2()
        { 
            // Sample 2
            MyDerivedClass d = new MyDerivedClass();
            d.Message(); // prints "MyDerivedClass".
            IMessage m = d as IMessage;
            m.Message(); // prints " MyDerivedClass "
        }

        public static void Test3()
        {
            MyDerivedClass d = new MyDerivedClass();
            d.Message(); // prints "MyDerivedClass".
            IMessage m = d as IMessage;
            m.Message(); // prints "MyDerivedClass"
            MyClass b = d;
            b.Message(); // prints "MyClass"

        }
    }

    namespace Fixes
    {
        public class MyDerivedClass : MyClass
        {
            public new void Message() =>
                WriteLine(nameof(MyDerivedClass));
        }

    }

    namespace PartThree
    {
        public class MyClass : IMessage
        {
            public virtual void Message() =>
                WriteLine(nameof(MyClass));
        }

        public class MyDerivedClass : MyClass
        {
            public override void Message() =>
                WriteLine(nameof(MyDerivedClass));
        }

    }

    namespace Partfour
    {
        public abstract class MyClass : IMessage
        {
            public abstract void Message();
        }

        public class MyDerivedClass : MyClass
        {
            public sealed override void Message() =>
                WriteLine(nameof(MyDerivedClass));
        }
    }

    namespace PartFive
    {
        public class MyClass : IMessage
        {
            protected virtual void OnMessage()
            {
            }

            public void Message()
            {
                OnMessage();
                WriteLine(nameof(MyClass));
            }
        }

        public class DefaultMessageGenerator
        {
            public void Message() =>
                WriteLine("This is a default message");
        }

        public class AnotherMessageGenerator :
            DefaultMessageGenerator, IMessage
        {
            // No explicit Message() method needed.
        }

    }
}
