﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Chapter2.MethodGroups
{
    public class Vector
    {
        private List<double> values = new List<double>();

        // Add a value to the internal list.
        public void Add(double number) =>
            values.Add(number);

        // Add values to each item in the sequence.
        public void Add(IEnumerable<double> sequence)
        {
            int index = 0;
            foreach (double number in sequence)
            {
                if (index == values.Count)
                    return;
                values[index++] += number;
            }
        }

        public void Scale(short scaleFactor)
        {
            for (int index = 0; index < values.Count; index++)
                values[index] *= scaleFactor;
        }

        public void Scale(int scaleFactor)
        {
            for (int index = 0; index < values.Count; index++)
                values[index] *= scaleFactor;
        }

        public void Scale(float scaleFactor)
        {
            for (int index = 0; index < values.Count; index++)
                values[index] *= scaleFactor;
        }

        public void Scale(double scaleFactor)
        {
            for (int index = 0; index < values.Count; index++)
                values[index] *= scaleFactor;
        }

        // Reuse the last 2 for the third snippet

    }

    public class Point
    {
        public double X { get; set; }
        public double Y { get; set; }

        public void Scale(int xScale, int yScale)
        {
            X *= xScale;
            Y *= yScale;
        }

        public void Scale(double xScale, double yScale)
        {
            X *= xScale;
            Y *= yScale;
        }
    }


    class Item22
    {
        public static void Tests()
        {
            Point p = new Point { X = 5, Y = 7 };
            // Note that second parameter is a long:
            p.Scale(5, 7L); // calls Scale(double,double)

        }
    }

    namespace PartTwo
    {
        public class Point
        {
            public double X { get; set; }
            public double Y { get; set; }

            // earlier code elided
            public void Scale(int scale)
            {
                X *= scale;
                Y *= scale;
            }
        }
        public class Point3D : Point
        {
            public double Z
            {
                get;
                set;
            }

            // Not override, not new. Different parameter type.
            public void Scale(double scale)
            {
                X *= scale;
                Y *= scale;
                Z *= scale;
            }
        }

        public class Testing
        {
            public static void Tests()
            {

                Point3D p2 = new Point3D { X = 1, Y = 2, Z = 3 };
                p2.Scale(3);

                double a1 = Utilities.Max(1, 3);
                double a2 = Utilities.Max(5.3, 12.7f);
                double a3 = Utilities.Max(5, 12.7f);

            }
        }

        public static class Utilities
        {
            // Prefer Math.Max for double:
            public static double Max(double left, double right)
            {
                return Math.Max(left, right);
            }

            // Note that float, int, etc. are handled here:
            public static T Max<T>(T left, T right) where T : IComparable<T>
            {
                return (left.CompareTo(right) > 0 ? left : right);
            }
        }

    }
}
