﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace Chapter3_Task
{
    class Item30
    {
        public static void Tests()
        {

        }

        public static async Task<IEnumerable<StockResult>> 
            ReadStockTickerV1(IEnumerable<string> symbols)
        {
            var results = new List<StockResult>();
            foreach (var symbol in symbols)
            {
                var result = await ReadSymbol(symbol);
                results.Add(result);
            }
            return results;
        }

        public static async Task<IEnumerable<StockResult>>
            ReadStockTickerV2(IEnumerable<string> symbols)
        {
            var resultTasks = new List<Task<StockResult>>();
            foreach (var symbol in symbols)
            {
                resultTasks.Add(ReadSymbol(symbol));
            }
            var results = await Task.WhenAll(resultTasks);

            return results.OrderBy(s => s.Price);
        }

        public static async Task<StockResult>
            ReadStockTickerV3(string symbol, IEnumerable<string> sources)
        {
            var resultTasks = new List<Task<StockResult>>();
            foreach (var source in sources)
            {
                resultTasks.Add(ReadSymbol(symbol, source));
            }
            var first = await Task.WhenAny(resultTasks);
            return await first;
        }

        public static async Task<IEnumerable<StockResult>>
            ReadStockTicker(IEnumerable<string> symbols)
        {
            var resultTasks = new List<Task<StockResult>>();
            var results = new List<StockResult>();
            foreach (var symbol in symbols)
            {
                resultTasks.Add(ReadSymbol(symbol));
            }
            foreach(var task in resultTasks)
            {
                var result = await task;
                results.Add(result);
            }
            return results;
        }

        public static async Task<IEnumerable<StockResult>>
            ReadStockTickerV5(IEnumerable<string> symbols)
        {
            var resultTasks = new List<Task<StockResult>>();
            var results = new List<StockResult>();
            foreach (var symbol in symbols)
            {
                resultTasks.Add(ReadSymbol(symbol));
            }
            while (resultTasks.Any())
            {
                // Each time through the loop, this creates a 
                // new task. That can be expensive.
                Task<StockResult> finishedTask = await Task.WhenAny(resultTasks);
                var result = await finishedTask;
                resultTasks.Remove(finishedTask);
                results.Add(result);
            }
            return results;
        }

        private static Task<StockResult> ReadSymbol(string symbol, string source)
        {
            return Task.FromResult(new StockResult());
        }
        private static Task<StockResult> ReadSymbol(string symbol)
        {
            return Task.FromResult(new StockResult());
        }
    }

    public class StockResult
    {
        public object Price { get; internal set; }
    }

    public static class Utilities2
    {
        public static Task<T>[] OrderByCompletion<T>(this IEnumerable<Task<T>> tasks)
        {
            // Copy to List because it gets enumerated multiple times.
            var sourceTasks = tasks.ToList();

            // Allocate the sources, allocate the output tasks.
            // Each output task is the corresponding task from
            // each completion source.
            var completionSources = new TaskCompletionSource<T>[sourceTasks.Count];
            var outputTasks = new Task<T>[completionSources.Length];
            for (int i = 0; i < completionSources.Length; i++)
            {
                completionSources[i] = new TaskCompletionSource<T>();
                outputTasks[i] = completionSources[i].Task;
            }

            // Magic, part one:
            // Each task has a continuation that puts its 
            // result in the next open location in the completion
            // sources array.
            int nextTaskIndex = -1;
            Action<Task<T>> continuation = completed =>
            {
                var bucket = completionSources[Interlocked.Increment(ref nextTaskIndex)];
                if (completed.IsCompleted)
                    bucket.TrySetResult(completed.Result);
                else if (completed.IsFaulted)
                    bucket.TrySetException(completed.Exception);
            };

            // Magic, part two:
            // for each input task, configure the 
            // continuation to set the output task
            // As each task completes, it uses the next location.
            foreach (var inputTask in sourceTasks)
            {
                inputTask.ContinueWith(continuation, CancellationToken.None,
                    TaskContinuationOptions.ExecuteSynchronously, TaskScheduler.Default);
            }

            return outputTasks;
        }

    }
}
