﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Text;
using System.Threading.Tasks;

namespace Chapter3
{
    public class Item26
    {

        public static void Tests()
        {
            var urls = new List<string>{"http://www.microsoft.com",
                "http://www.google.com",
                "http://www.cnn.com",
                "http://www.srtsolutions.com",
                "http://doesnotexist.srtsolutions.com",
                "http://www.srtsolutions.com/blogs/billwagner",
                "http://www.espn.com",
                "http://www.uiuc.edu"};

            try
            {
                urls.RunAsync(
                    url => startDownload(url),
                    task => finishDownload(
                        task.AsyncState.ToString(), task.Result));
            }
            catch (AggregateException problems)
            {
                ReportAggregateError(problems);
            }

            try
            {
                urls.RunAsync(
                    url => startDownload(url),
                    task => finishDownload(task.AsyncState.ToString(),
                    task.Result));
            }
            catch (AggregateException problems)
            {
                var handlers = new Dictionary<Type, Action<Exception>>();
                handlers.Add(typeof(WebException),
                    ex => Console.WriteLine(ex.Message));

                if (!HandleAggregateError(problems, handlers))
                    throw;
            }

            var data = Enumerable.Range(1, 50);

            var nums = from n in data
                       where n < 150
                       select Factorial(n);

            try
            {
                foreach (var item in nums)
                    Console.WriteLine(item);
            }
            catch (InvalidOperationException inv)
            {
                // elided
            }


        }

        private static int Factorial(int num)
        {
            int factImpl(int n)
            {
                return n == 1 ? 1 : n * factImpl(n - 1);
            }
            return factImpl(num);

        }
        private static bool HandleAggregateError(AggregateException aggregate,
            Dictionary<Type, Action<Exception>> exceptionHandlers)
        {
            foreach (var exception in aggregate.InnerExceptions)
            {
                if (exception is AggregateException agEx)
                {
                    if (!HandleAggregateError(agEx, exceptionHandlers))
                    {
                        return false;
                    } else
                    {
                        continue;
                    }
                }
                else if (exceptionHandlers.ContainsKey(exception.GetType()))
                {
                    exceptionHandlers[exception.GetType()](exception);
                }
                else
                    return false;
            }
            return true;
        }


        private static void ReportAggregateError(
            AggregateException aggregate)
        {
            foreach (var exception in aggregate.InnerExceptions)
                if (exception is AggregateException agEx)
                    ReportAggregateError(agEx);
                else
                    Console.WriteLine(exception.Message);
        }

        private static void finishDownload(string url, byte[] bytes)
        {
            Console.WriteLine("Read {0} bytes from {1}",
                bytes.Length, url);
        }


        private static Task<byte[]> startDownload(string url)
        {
            var tcs = new TaskCompletionSource<byte[]>(url);
            var wc = new WebClient();
            wc.DownloadDataCompleted += (sender, e) =>
            {
                if (e.UserState == tcs)
                {
                    if (e.Cancelled)
                        tcs.TrySetCanceled();
                    else if (e.Error != null)
                    {
                        if (e.Error is WebException)
                            tcs.TrySetResult(new byte[0]);
                        else
                            tcs.TrySetException(e.Error);
                    }
                    else
                        tcs.TrySetResult(e.Result);
                }
            };
            wc.DownloadDataAsync(new Uri(url), tcs);
            return tcs.Task;
        }


    }

    public static class Extensions
    {
        public static void RunAsync<T, TResult>(
            this IEnumerable<T> taskParms,
            Func<T, Task<TResult>> taskStarter,
            Action<Task<TResult>> taskFinisher)
        {
            taskParms.Select(parm => taskStarter(parm)).
                AsParallel().
                ForAll(t => t.ContinueWith(t2 => taskFinisher(t2)));
        }

    }
}
