﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace Chapter3
{
    public class Item30
    {
        private int total = 0;
        private object syncHandle = new object();

        public int TotalValue
        {
            get
            {
                lock (syncHandle) { return total; }
            }
        }

        public void IncrementTotal()
        {
            lock (syncHandle) { total++; }
        }

        public void IncrementTotal2()
        {
            object tmpObject = syncHandle;
            System.Threading.Monitor.Enter(tmpObject);
            try
            {
                total++;
            }
            finally
            {
                System.Threading.Monitor.Exit(tmpObject);
            }
        }

        /*
        public void IncrementTotal()
        {
            lock (total) // compiler error: can't lock value type
            {
                total++;
            }
        }

        public void IncrementTotal()
        {
            // really doesn’t lock total.
            // locks a box containing total.
            Monitor.Enter(total);
            try
            {
                total++;
            }
            finally
            {
                // Might throw exception
                // unlocks a different box containing total
                Monitor.Exit(total);
            }
        }
        public void IncrementTotal()
        {
            // doesn’t work either:
            object lockHandle = total;
            Monitor.Enter(lockHandle);
            try
            {
                total++;
            }
            finally
            {
                Monitor.Exit(lockHandle);
            }
        }

        public void IncrementTotal()
        {
            if (!Monitor.TryEnter(syncHandle, 1000)) // wait 1 second
                throw new PreciousResourceException
                    ("Could not enter critical section");
            try
            {
                total++;
            }
            finally
            {
                Monitor.Exit(syncHandle);
            }
        }

        public void IncrementTotal() =>
            total++;

        public void IncrementTotal() =>
            System.Threading.Interlocked.Increment(ref total);
        */

        public void IncrementSomething()
        {
            object lockHandle = new object();

            using (LockHolder<object> lockObj = new LockHolder<object>
                (lockHandle, 1000))
            {
                if (lockObj.LockSuccessful)
                {
                    // work elided
                }
            }
            // Dispose called here.
        }
    }

    public sealed class LockHolder<T> : IDisposable
        where T : class
    {
        private T handle;
        private bool holdsLock;

        public LockHolder(T handle, int milliSecondTimeout)
        {
            this.handle = handle;
            holdsLock = System.Threading.Monitor.TryEnter(
                handle, milliSecondTimeout);
        }

        public bool LockSuccessful
        {
            get { return holdsLock; }
        }

        public void Dispose()
        {
            if (holdsLock)
                System.Threading.Monitor.Exit(handle);
            // Don't unlock twice
            holdsLock = false;
        }
    }

}
