﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;

namespace Chapter4
{
    public class Item40
    {
        public static dynamic Add(dynamic left,
            dynamic right)
        {
            return left + right;
        }

        public static void SnippetOne()
        {
            dynamic answer = Add(5, 5);
            answer = Add(5.5, 7.3);
            answer = Add(5, 12.3);
        }

        public static void SnippetTwo_Three()
        {
            dynamic label = Add("Here is ", "a label");

            dynamic tomorrow = Add(DateTime.Now, TimeSpan.FromDays(1));

        }

        public static void SnippetFour()
        {
            dynamic answer = Add(5, 12.3);
            int value = (int)answer;
            string stringLabel = System.Convert.ToString(answer);
        }

        public static TResult Add<T1, T2, TResult>(T1 left, T2 right,
            Func<T1, T2, TResult> AddMethod)
        {
            return AddMethod(left, right);
        }

        public static void SnippetFive()
        {
            var lambdaAnswer = Add(5, 5, (a, b) => a + b);
            var lambdaAnswer2 = Add(5.5, 7.3, (a, b) => a + b);
            var lambdaAnswer3 = Add(5, 12.3, (a, b) => a + b);
            var lambdaLabel = Add("Here is ", "a label",
                (a, b) => a + b);
            dynamic tomorrow = Add(DateTime.Now, TimeSpan.FromDays(1));
            var finalLabel = Add("something", 3,
                (a, b) => a + b.ToString());
        }

        public static void SnippetSix()
        {
            var sequence = Enumerable.Range(1, 50);

            var accumulatedTotal = Enumerable.Aggregate(sequence,
                (a, b) => a + b);
        }

        // Naive Implementation. Read on for a better version
        public static T AddExpression<T>(T left, T right)
        {
            ParameterExpression leftOperand = Expression.Parameter(
                typeof(T), "left");
            ParameterExpression rightOperand = Expression.Parameter(
                typeof(T), "right");
            BinaryExpression body = Expression.Add(
                leftOperand, rightOperand);
            Expression<Func<T, T, T>> adder =
                Expression.Lambda<Func<T, T, T>>(
                body, leftOperand, rightOperand);
            Func<T, T, T> theDelegate = adder.Compile();
            return theDelegate(left, right);
        }

        public static void SnippetSeven()
        {
            int sum = AddExpression(5, 7);
        }

        private static class VersionTwo
        {
            // A little better.
            public static TResult AddExpression<T1, T2, TResult>
                (T1 left, T2 right)
            {
                var leftOperand = Expression.Parameter(typeof(T1),
                    "left");
                var rightOperand = Expression.Parameter(typeof(T2),
                    "right");
                var body = Expression.Add(leftOperand, rightOperand);
                var adder = Expression.Lambda<Func<T1, T2, TResult>>(
                    body, leftOperand, rightOperand);
                return adder.Compile()(left, right);
            }

            public static void Testing()
            {
                int sum2 = AddExpression<int, int, int>(5, 7);

                DateTime nextWeek = AddExpression<DateTime, TimeSpan,
                    DateTime>(
                    DateTime.Now, TimeSpan.FromDays(7));

            }
        }

        // dangerous but working version
        public static class BinaryOperator<T1, T2, TResult>
        {
            static Func<T1, T2, TResult> compiledExpression;

            public static TResult Add(T1 left, T2 right)
            {
                if (compiledExpression == null)
                    createFunc();

                return compiledExpression(left, right);
            }

            private static void createFunc()
            {
                var leftOperand = Expression.Parameter(typeof(T1),
                    "left");
                var rightOperand = Expression.Parameter(typeof(T2),
                    "right");
                var body = Expression.Add(leftOperand, rightOperand);
                var adder = Expression.Lambda<Func<T1, T2, TResult>>(
                    body, leftOperand, rightOperand);
                compiledExpression = adder.Compile();
            }
        }

        // A fix for one problem causes another
        public static TResult AddExpressionWithConversion
            <T1, T2, TResult>(T1 left, T2 right)
        {
            var leftOperand = Expression.Parameter(typeof(T1),
                "left");
            Expression convertedLeft = leftOperand;
            if (typeof(T1) != typeof(TResult))
            {
                convertedLeft = Expression.Convert(leftOperand,
                    typeof(TResult));
            }
            var rightOperand = Expression.Parameter(typeof(T2),
                "right");
            Expression convertedRight = rightOperand;
            if (typeof(T2) != typeof(TResult))
            {
                convertedRight = Expression.Convert(rightOperand,
                typeof(TResult));
            }
            var body = Expression.Add(convertedLeft, convertedRight);
            var adder = Expression.Lambda<Func<T1, T2, TResult>>(
                body, leftOperand, rightOperand);
            return adder.Compile()(left, right);
        }

        public static class BinaryOperators<T>
        {
            static Func<T, T, T> compiledExpression;

            public static T Add(T left, T right)
            {
                if (compiledExpression == null)
                    createFunc();

                return compiledExpression(left, right);
            }

            private static void createFunc()
            {
                var leftOperand = Expression.Parameter(typeof(T),
                    "left");
                var rightOperand = Expression.Parameter(typeof(T),
                    "right");
                var body = Expression.Add(leftOperand, rightOperand);
                var adder = Expression.Lambda<Func<T, T, T>>(
                    body, leftOperand, rightOperand);
                compiledExpression = adder.Compile();
            }
        }

    }
}
